deferred class
	CE_CLASS

inherit
	INTERNAL
		rename
			class_name as class_name_from_eif_internal,
			field as field_from_eif_internal,
			mark as mark_from_eif_internal,
			type_name as type_name_from_eif_internal,
			unmark as unmark_from_eif_internal
		end
	DISPOSABLE
		rename
			dispose as dispose_from_eif_disposable
		end

feature {CE_CLASS, CE_POINTER, CE_VA_LIST}

	make
		do
		end

	make_memory
		local
			l_fields: like structure_info.fields
			l_i: INTEGER_32
			l_field: CE_STRUCTURE_FIELD
			l_old_state: BOOLEAN
			l_tmp_array: CE_ARRAY [ANY]
			l_tmp_class: CE_CLASS
			l_tmp_pointer: CE_POINTER [ANY]
		do
			internal_c_pointer := default_pointer.memory_calloc (1, structure_size)
			is_allocated_memory := True
			from
				l_fields := structure_info.fields
				l_i := l_fields.count - 1
			until
				l_i < 0
			loop
				l_field := l_fields[l_i]
				inspect l_field.ce_type
				when {CE_UTILITIES}.Ce_type_array then
					l_old_state := {ISE_RUNTIME}.check_assert (False)
					l_tmp_array ?= new_instance_of (l_field.eif_type)
					l_tmp_array.make_shared (internal_c_pointer + l_field.c_offset, l_field.dimensions)
					l_old_state := {ISE_RUNTIME}.check_assert (l_old_state)
					set_reference_field (l_field.internal_index, Current, l_tmp_array)
				when {CE_UTILITIES}.Ce_type_class then
					l_old_state := {ISE_RUNTIME}.check_assert (False)
					l_tmp_class ?= new_instance_of (l_field.eif_type)
					l_tmp_class.make_shared (internal_c_pointer + l_field.c_offset)
					l_old_state := {ISE_RUNTIME}.check_assert (l_old_state)
					set_reference_field (l_field.internal_index, Current, l_tmp_class)
				when {CE_UTILITIES}.Ce_type_pointer then
					l_tmp_pointer ?= new_instance_of (l_field.eif_type)
					set_reference_field (l_field.internal_index, Current, l_tmp_pointer)
				else
					-- nothing to do
				end
				l_i := l_i - 1
			end
		end

	make_shared (a_c_pointer: POINTER)
		local
			l_fields: like structure_info.fields
			l_i: INTEGER_32
			l_field: CE_STRUCTURE_FIELD
			l_old_state: BOOLEAN
			l_tmp_array: CE_ARRAY [ANY]
			l_tmp_class: CE_CLASS
		do
			internal_c_pointer := a_c_pointer
			from
				l_fields := structure_info.fields
				l_i := l_fields.count - 1
			until
				l_i < 0
			loop
				l_field := l_fields[l_i]
				inspect l_field.ce_type
				when {CE_UTILITIES}.Ce_type_array then
					l_old_state := {ISE_RUNTIME}.check_assert (False)
					l_tmp_array ?= new_instance_of (l_field.eif_type)
					l_tmp_array.make_shared (a_c_pointer + l_field.c_offset, l_field.dimensions)
					l_old_state := {ISE_RUNTIME}.check_assert (l_old_state)
					set_reference_field (l_field.internal_index, Current, l_tmp_array)
				when {CE_UTILITIES}.Ce_type_class then
					l_old_state := {ISE_RUNTIME}.check_assert (False)
					l_tmp_class ?= new_instance_of (l_field.eif_type)
					l_tmp_class.make_shared (a_c_pointer + l_field.c_offset)
					l_old_state := {ISE_RUNTIME}.check_assert (l_old_state)
					set_reference_field (l_field.internal_index, Current, l_tmp_class)
				else
					update_class (l_field)
				end
				l_i := l_i - 1
			end
		end

feature {NONE}

	make_initialized (a_data: TUPLE)
		require
			a_data /= Void and then a_data.count = structure_info.fields.count
			across structure_info.fields as c all
					type_conforms_to (a_data[c.target_index + 1].generating_type.type_id, c.item.eif_type)
				end
		do
			across structure_info.fields as c loop
				inspect c.item.ce_type
				when {CE_UTILITIES}.Ce_type_boolean    then set_boolean_field    (c.item.internal_index, Current, a_data.boolean_item    (c.target_index + 1))
				when {CE_UTILITIES}.Ce_type_integer_8  then set_integer_8_field  (c.item.internal_index, Current, a_data.integer_8_item  (c.target_index + 1))
				when {CE_UTILITIES}.Ce_type_integer_16 then set_integer_16_field (c.item.internal_index, Current, a_data.integer_16_item (c.target_index + 1))
				when {CE_UTILITIES}.Ce_type_integer_32 then set_integer_32_field (c.item.internal_index, Current, a_data.integer_32_item (c.target_index + 1))
				when {CE_UTILITIES}.Ce_type_integer_64 then set_integer_64_field (c.item.internal_index, Current, a_data.integer_64_item (c.target_index + 1))
				when {CE_UTILITIES}.Ce_type_natural_8  then set_natural_8_field  (c.item.internal_index, Current, a_data.natural_8_item  (c.target_index + 1))
				when {CE_UTILITIES}.Ce_type_natural_16 then set_natural_16_field (c.item.internal_index, Current, a_data.natural_16_item (c.target_index + 1))
				when {CE_UTILITIES}.Ce_type_natural_32 then set_natural_32_field (c.item.internal_index, Current, a_data.natural_32_item (c.target_index + 1))
				when {CE_UTILITIES}.Ce_type_natural_64 then set_natural_64_field (c.item.internal_index, Current, a_data.natural_64_item (c.target_index + 1))
				when {CE_UTILITIES}.Ce_type_real_32    then set_real_32_field    (c.item.internal_index, Current, a_data.real_32_item    (c.target_index + 1))
				when {CE_UTILITIES}.Ce_type_real_64    then set_real_64_field    (c.item.internal_index, Current, a_data.real_64_item    (c.target_index + 1))
				when {CE_UTILITIES}.Ce_type_real_96 then
					if
						attached {REAL_96} a_data.reference_item (c.target_index + 1) as init_r96 and
						attached {REAL_96} field_from_eif_internal (c.item.internal_index, Current) as class_r96
					then
						class_r96.c_pointer.memory_copy (init_r96.c_pointer, {CE_UTILITIES}.Real_96_bytes)
					else
						(create {EXCEPTIONS}).raise (
								a_data.reference_item (c.target_index + 1).generator +
								" <-> " +
								field_from_eif_internal (c.item.internal_index, Current).generator
							)
					end
				else
					set_reference_field (c.item.internal_index, Current, a_data.reference_item (c.target_index + 1))
				end
			end
			if structure_info.is_union or structure_info.has_bitfields then
				create_memory_copy
			end
		end

feature {CE_CLASS} -- data

	internal_c_pointer: POINTER

	is_allocated_memory: BOOLEAN

	dispose_from_eif_disposable
		do
			if is_allocated_memory then
				internal_c_pointer.memory_free
			end
		end

feature {ANY}

	create_memory_copy
		local
			l_fields: like structure_info.fields
			l_i: INTEGER_32
			l_field: CE_STRUCTURE_FIELD
			l_old_state: BOOLEAN
			l_tmp_array: CE_ARRAY [ANY]
		do
			if internal_c_pointer = default_pointer then
				internal_c_pointer := default_pointer.memory_calloc (1, structure_size)
				is_allocated_memory := True

				from
					l_fields := structure_info.fields
					l_i := l_fields.count - 1
				until
					l_i < 0
				loop
					l_field := l_fields[l_i]
					inspect l_field.ce_type
					when {CE_UTILITIES}.Ce_type_array then
						l_tmp_array ?= field_from_eif_internal (l_field.internal_index, Current)
						(internal_c_pointer + l_field.c_offset).memory_copy (l_tmp_array.c_pointer, l_tmp_array.total_count * l_tmp_array.element_size)

						l_old_state := {ISE_RUNTIME}.check_assert (False)
						l_tmp_array ?= new_instance_of (l_field.eif_type)
						l_tmp_array.make_shared (internal_c_pointer + l_field.c_offset, l_field.dimensions)
						l_old_state := {ISE_RUNTIME}.check_assert (l_old_state)
						set_reference_field (l_field.internal_index, Current, l_tmp_array)
					else
						update_memory (l_field)
					end
					l_i := l_i - 1
				end
			end
		end

feature {ANY}

	address: CE_POINTER [like Current]
		do
			create {CE_CLASS_POINTER [like Current]} Result.make_shared (c_pointer)
		end

	c_pointer: POINTER
		do
			create_memory_copy
			Result := internal_c_pointer
		end

	offset_of (a_field_name: STRING_8): INTEGER_32
		require
			a_field_name /= Void
		do
			Result := structure_info.field_table[a_field_name].c_offset
		end

	structure_size: INTEGER_32
		do
			Result := structure_info.size
		end

	update_memory_field (a_name: STRING_8)
		require
			a_name /= Void
		local
			l_field: CE_STRUCTURE_FIELD
		do
			if internal_c_pointer /= default_pointer then
				l_field := structure_info.field_table[a_name]
				update_memory (l_field)
				if l_field.is_bitfield then
					update_class (l_field)
				end
			end
		end

	update_memory_fields
		local
			l_fields: like structure_info.fields
			l_i: INTEGER_32
		do
			from
				l_fields := structure_info.fields
				l_i := l_fields.count - 1
			until
				l_i < 0
			loop
				update_memory (l_fields[l_i])
				l_i := l_i - 1
			end
		end

	update_class_field (a_name: STRING_8)
		require
			a_name /= Void
		do
			update_class (structure_info.field_table[a_name])
		end

	update_class_fields
		local
			l_fields: like structure_info.fields
			l_i: INTEGER_32
		do
			from
				l_fields := structure_info.fields
				l_i := l_fields.count - 1
			until
				l_i < 0
			loop
				update_class (l_fields[l_i])
				l_i := l_i - 1
			end
		end

feature {CE_CLASS, CE_STRUCTURE, CE_UTILITIES}

	structure_info: CE_STRUCTURE
		deferred
		end

feature {NONE}

	init_structure_info
		local
			l_type_id, l_i, l_count: INTEGER_32
			l_internal_info: HASH_TABLE [TUPLE [type: TYPE [ANY]; internal_index: INTEGER_32], STRING] -- field name is lower case
			l_type: TYPE [ANY]
			l_size, l_c_offset, l_alignment: INTEGER_32
			l_old_state: BOOLEAN
			l_tmp_array: CE_ARRAY [ANY]
			l_tmp_class: CE_CLASS
		do
			-- get the INTERNAL info
			from
				l_type_id := dynamic_type (Current)
				l_i := 1
				l_count := field_count_of_type (l_type_id)
				create l_internal_info.make (l_count)
			until
				l_i > l_count
			loop
				l_internal_info[field_name_of_type (l_i, l_type_id)] := [type_of_type (field_static_type_of_type (l_i, l_type_id)), l_i]
				l_i := l_i + 1
			end

			-- fill in data
			across structure_info.fields as c loop
				l_type := l_internal_info[c.item.name.as_lower].type
				c.item.ce_type := (create {CE_UTILITIES}).ce_type (l_type)
				if attached {TYPE [CE_POINTER [ANY]]} l_type and then not attached {TYPE [CE_ROUTINE]} l_type then
					c.item.eif_type := (create {CE_UTILITIES}).effective_type_id (
							not attached {TYPE [CE_ARRAY [ANY]]} l_type,
							l_type.generic_parameter_type (1)
						)
				else
					c.item.eif_type := l_type.type_id
				end
				c.item.internal_index := l_internal_info[c.item.name.as_lower].internal_index

				inspect c.item.ce_type
				when {CE_UTILITIES}.Ce_type_boolean    then l_size := {PLATFORM}.Boolean_bytes
				when {CE_UTILITIES}.Ce_type_integer_8  then l_size := {PLATFORM}.Integer_8_bytes
				when {CE_UTILITIES}.Ce_type_integer_16 then l_size := {PLATFORM}.Integer_16_bytes
				when {CE_UTILITIES}.Ce_type_integer_32 then l_size := {PLATFORM}.Integer_32_bytes
				when {CE_UTILITIES}.Ce_type_integer_64 then l_size := {PLATFORM}.Integer_64_bytes
				when {CE_UTILITIES}.Ce_type_natural_8  then l_size := {PLATFORM}.Natural_8_bytes
				when {CE_UTILITIES}.Ce_type_natural_16 then l_size := {PLATFORM}.Natural_16_bytes
				when {CE_UTILITIES}.Ce_type_natural_32 then l_size := {PLATFORM}.Natural_32_bytes
				when {CE_UTILITIES}.Ce_type_natural_64 then l_size := {PLATFORM}.Natural_64_bytes
				when {CE_UTILITIES}.Ce_type_real_32    then l_size := {PLATFORM}.Real_32_bytes
				when {CE_UTILITIES}.Ce_type_real_64    then l_size := {PLATFORM}.Real_64_bytes
				when {CE_UTILITIES}.Ce_type_real_96    then l_size := {CE_UTILITIES}.Real_96_bytes
				when {CE_UTILITIES}.Ce_type_va_list    then l_size := {PLATFORM}.Pointer_bytes
				when {CE_UTILITIES}.Ce_type_pointer    then l_size := {PLATFORM}.Pointer_bytes
				when {CE_UTILITIES}.Ce_type_array then
					l_old_state := {ISE_RUNTIME}.check_assert (False)
					l_tmp_array ?= new_instance_of (c.item.eif_type)
					l_size := l_tmp_array.element_size
					if l_size = 0 then -- CE_ARRAY_ARRAY
						l_tmp_array.make_shared (default_pointer + 1, <<0>>)
						l_size := l_tmp_array.element_size
					end
					l_old_state := {ISE_RUNTIME}.check_assert (l_old_state)

					across c.item.dimensions as c2 loop l_size := l_size * c2.item end
				when {CE_UTILITIES}.Ce_type_class then
					l_old_state := {ISE_RUNTIME}.check_assert (False)
					l_tmp_class ?= new_instance_of (c.item.eif_type)
					l_size :=  l_tmp_class.structure_size
					l_old_state := {ISE_RUNTIME}.check_assert (l_old_state)
				end

				if c.item.is_bitfield then
					if structure_info.is_union then
						structure_info.size := structure_info.size.max (l_size)
					elseif c.target_index + 1 = structure_info.fields.count then
						l_c_offset := bitfield_helper (1, default_pointer, default_pointer, 0).to_integer_32
					elseif
						attached structure_info.fields[c.target_index + 1] as next_field and then
						not next_field.is_bitfield and then
						attached (create {C_STRING}.make (next_field.name)) as c_name -- "local" variable
					then
						l_c_offset := bitfield_helper (1, c_name.item, default_pointer, 0).to_integer_32
					end
				elseif structure_info.is_struct then
					if c.item.alignment /= 0 then
						l_alignment := c.item.alignment
					else
						l_alignment := (create {CE_UTILITIES}).ce_alignof (l_type).to_integer_32
					end
					l_c_offset := (l_c_offset / l_alignment).ceiling * l_alignment
					c.item.c_offset := l_c_offset
					l_c_offset := l_c_offset + l_size
				else -- is_union
					structure_info.size := structure_info.size.max (l_size)
				end
			end
			if structure_info.is_struct then
				l_alignment := max_alignment (structure_info)
				structure_info.size := (l_c_offset / l_alignment).ceiling * l_alignment
			else
				-- unions handled above
			end
		end

	bitfield_helper (a_operation: INTEGER_32; a_field_name, a_struct_pointer: POINTER; a_value: NATURAL_64): NATURAL_64
			-- a_operation:
			--    1: get non-bitfield offset
			--    2: set bitfield
			--    3: get bitfield
		require
			a_operation = 1 or a_operation = 2 or a_operation = 3
			a_operation = 1 implies
				(
					a_field_name = default_pointer or else
					(attached structure_info.field_table[create {STRING_8}.make_from_c (a_field_name)] as f and then not f.is_bitfield)
				) and
				a_struct_pointer = default_pointer and
				a_value = 0
			a_operation = 2 implies
				(attached structure_info.field_table[create {STRING_8}.make_from_c (a_field_name)] as f and then f.is_bitfield) and
				a_struct_pointer /= default_pointer
			a_operation = 3 implies
				(attached structure_info.field_table[create {STRING_8}.make_from_c (a_field_name)] as f and then f.is_bitfield) and
				a_struct_pointer /= default_pointer and
				a_value = 0
		do
		end

	max_alignment (a_structure_info: CE_STRUCTURE): INTEGER_32
		require
			a_structure_info /= Void
		local
			l_tmp_class: CE_CLASS
			l_alignment: INTEGER_32
		do
			across a_structure_info.fields as c loop
				if c.item.eif_type = {CE_UTILITIES}.Ce_type_class then
					l_tmp_class ?= new_instance_of (c.item.ce_type)
					Result := Result.max (max_alignment (l_tmp_class.structure_info))
				else
					if c.item.alignment /= 0 then
						l_alignment := c.item.alignment
					else
						l_alignment := (create {CE_UTILITIES}).ce_alignof (type_of_type (c.item.eif_type)).to_integer_32
					end
					Result := Result.max (l_alignment)
				end
			end
		ensure
			Result = 0 or Result = 1 or Result = 2 or Result = 4
		end

feature {NONE} -- data synchronization

	update_memory (a_field: CE_STRUCTURE_FIELD)
		local
			l_c_name: C_STRING
			l_tmp_pointer: CE_POINTER [ANY]
			l_old_state: BOOLEAN
			l_tmp_class: CE_CLASS
			l_tmp_va_list: CE_VA_LIST
		do
			if a_field.is_bitfield then
				create l_c_name.make (a_field.name)
				inspect a_field.ce_type
				when {CE_UTILITIES}.Ce_type_boolean    then bitfield_helper (2, l_c_name.item, internal_c_pointer, boolean_field    (a_field.internal_index, Current).to_integer.to_natural_64).do_nothing
				when {CE_UTILITIES}.Ce_type_integer_8  then bitfield_helper (2, l_c_name.item, internal_c_pointer, integer_8_field  (a_field.internal_index, Current).to_natural_64).do_nothing
				when {CE_UTILITIES}.Ce_type_integer_16 then bitfield_helper (2, l_c_name.item, internal_c_pointer, integer_16_field (a_field.internal_index, Current).to_natural_64).do_nothing
				when {CE_UTILITIES}.Ce_type_integer_32 then bitfield_helper (2, l_c_name.item, internal_c_pointer, integer_32_field (a_field.internal_index, Current).to_natural_64).do_nothing
				when {CE_UTILITIES}.Ce_type_integer_64 then bitfield_helper (2, l_c_name.item, internal_c_pointer, integer_64_field (a_field.internal_index, Current).to_natural_64).do_nothing
				when {CE_UTILITIES}.Ce_type_natural_8  then bitfield_helper (2, l_c_name.item, internal_c_pointer, natural_8_field  (a_field.internal_index, Current)).do_nothing
				when {CE_UTILITIES}.Ce_type_natural_16 then bitfield_helper (2, l_c_name.item, internal_c_pointer, natural_16_field (a_field.internal_index, Current)).do_nothing
				when {CE_UTILITIES}.Ce_type_natural_32 then bitfield_helper (2, l_c_name.item, internal_c_pointer, natural_32_field (a_field.internal_index, Current)).do_nothing
				when {CE_UTILITIES}.Ce_type_natural_64 then bitfield_helper (2, l_c_name.item, internal_c_pointer, natural_64_field (a_field.internal_index, Current)).do_nothing
				end
			else
				inspect a_field.ce_type
				when {CE_UTILITIES}.Ce_type_boolean    then {CE_BOOLEAN_POINTER}.c_write    (internal_c_pointer, a_field.c_offset, boolean_field    (a_field.internal_index, Current))
				when {CE_UTILITIES}.Ce_type_integer_8  then {CE_INTEGER_8_POINTER}.c_write  (internal_c_pointer, a_field.c_offset, integer_8_field  (a_field.internal_index, Current))
				when {CE_UTILITIES}.Ce_type_integer_16 then {CE_INTEGER_16_POINTER}.c_write (internal_c_pointer, a_field.c_offset, integer_16_field (a_field.internal_index, Current))
				when {CE_UTILITIES}.Ce_type_integer_32 then {CE_INTEGER_32_POINTER}.c_write (internal_c_pointer, a_field.c_offset, integer_32_field (a_field.internal_index, Current))
				when {CE_UTILITIES}.Ce_type_integer_64 then {CE_INTEGER_64_POINTER}.c_write (internal_c_pointer, a_field.c_offset, integer_64_field (a_field.internal_index, Current))
				when {CE_UTILITIES}.Ce_type_natural_8  then {CE_NATURAL_8_POINTER}.c_write  (internal_c_pointer, a_field.c_offset, natural_8_field  (a_field.internal_index, Current))
				when {CE_UTILITIES}.Ce_type_natural_16 then {CE_NATURAL_16_POINTER}.c_write (internal_c_pointer, a_field.c_offset, natural_16_field (a_field.internal_index, Current))
				when {CE_UTILITIES}.Ce_type_natural_32 then {CE_NATURAL_32_POINTER}.c_write (internal_c_pointer, a_field.c_offset, natural_32_field (a_field.internal_index, Current))
				when {CE_UTILITIES}.Ce_type_natural_64 then {CE_NATURAL_64_POINTER}.c_write (internal_c_pointer, a_field.c_offset, natural_64_field (a_field.internal_index, Current))
				when {CE_UTILITIES}.Ce_type_real_32    then {CE_REAL_32_POINTER}.c_write    (internal_c_pointer, a_field.c_offset, real_32_field    (a_field.internal_index, Current))
				when {CE_UTILITIES}.Ce_type_real_64    then {CE_REAL_64_POINTER}.c_write    (internal_c_pointer, a_field.c_offset, real_64_field    (a_field.internal_index, Current))
				when {CE_UTILITIES}.Ce_type_real_96 then
					if attached {REAL_96} field_from_eif_internal (a_field.internal_index, Current) as r96 then
						(internal_c_pointer + a_field.c_offset).memory_copy (r96.c_pointer, {CE_UTILITIES}.Real_96_bytes)
					else
						(create {EXCEPTIONS}).raise (field_from_eif_internal (a_field.internal_index, Current).generator)
					end
				when {CE_UTILITIES}.Ce_type_pointer then
					l_tmp_pointer ?= field_from_eif_internal (a_field.internal_index, Current)
					{CE_POINTER_POINTER [ANY]}.c_write (internal_c_pointer, a_field.c_offset, l_tmp_pointer.c_pointer)
				when {CE_UTILITIES}.Ce_type_array then
					-- nothing to do, uses the struct memory
				when {CE_UTILITIES}.Ce_type_class then
					l_tmp_class ?= field_from_eif_internal (a_field.internal_index, Current)
					if l_tmp_class.c_pointer /= internal_c_pointer + a_field.c_offset then
						(internal_c_pointer + a_field.c_offset).memory_copy (l_tmp_class.c_pointer, l_tmp_class.structure_size)

						l_old_state := {ISE_RUNTIME}.check_assert (False)
						l_tmp_class ?= new_instance_of (a_field.eif_type)
						l_tmp_class.make_shared (internal_c_pointer + a_field.c_offset)
						l_old_state := {ISE_RUNTIME}.check_assert (l_old_state)

						set_reference_field (a_field.internal_index, Current, l_tmp_class)
					end
				when {CE_UTILITIES}.Ce_type_va_list then
					l_tmp_va_list ?= field_from_eif_internal (a_field.internal_index, Current)
					{CE_POINTER_POINTER [ANY]}.c_write (internal_c_pointer, a_field.c_offset, l_tmp_va_list.c_pointer)
				end
			end
		end

	update_class (a_field: CE_STRUCTURE_FIELD)
		local
			l_c_name: C_STRING
			l_tmp_pointer: CE_POINTER [ANY]
			l_tmp_class: CE_CLASS
		do
			if a_field.is_bitfield then
				create l_c_name.make (a_field.name)
				inspect a_field.ce_type
				when {CE_UTILITIES}.Ce_type_boolean    then set_boolean_field    (a_field.internal_index, Current, bitfield_helper (3, l_c_name.item, internal_c_pointer, 0).to_boolean)
				when {CE_UTILITIES}.Ce_type_integer_8  then set_integer_8_field  (a_field.internal_index, Current, bitfield_helper (3, l_c_name.item, internal_c_pointer, 0).to_integer_8)
				when {CE_UTILITIES}.Ce_type_integer_16 then set_integer_16_field (a_field.internal_index, Current, bitfield_helper (3, l_c_name.item, internal_c_pointer, 0).to_integer_16)
				when {CE_UTILITIES}.Ce_type_integer_32 then set_integer_32_field (a_field.internal_index, Current, bitfield_helper (3, l_c_name.item, internal_c_pointer, 0).to_integer_32)
				when {CE_UTILITIES}.Ce_type_integer_64 then set_integer_64_field (a_field.internal_index, Current, bitfield_helper (3, l_c_name.item, internal_c_pointer, 0).to_integer_64)
				when {CE_UTILITIES}.Ce_type_natural_8  then set_natural_8_field  (a_field.internal_index, Current, bitfield_helper (3, l_c_name.item, internal_c_pointer, 0).to_natural_8)
				when {CE_UTILITIES}.Ce_type_natural_16 then set_natural_16_field (a_field.internal_index, Current, bitfield_helper (3, l_c_name.item, internal_c_pointer, 0).to_natural_16)
				when {CE_UTILITIES}.Ce_type_natural_32 then set_natural_32_field (a_field.internal_index, Current, bitfield_helper (3, l_c_name.item, internal_c_pointer, 0).to_natural_32)
				when {CE_UTILITIES}.Ce_type_natural_64 then set_natural_64_field (a_field.internal_index, Current, bitfield_helper (3, l_c_name.item, internal_c_pointer, 0).to_natural_64)
				end
			else
				inspect a_field.ce_type
				when {CE_UTILITIES}.Ce_type_boolean    then set_boolean_field    (a_field.internal_index, Current, {CE_BOOLEAN_POINTER}.c_read    (internal_c_pointer, a_field.c_offset))
				when {CE_UTILITIES}.Ce_type_integer_8  then set_integer_8_field  (a_field.internal_index, Current, {CE_INTEGER_8_POINTER}.c_read  (internal_c_pointer, a_field.c_offset))
				when {CE_UTILITIES}.Ce_type_integer_16 then set_integer_16_field (a_field.internal_index, Current, {CE_INTEGER_16_POINTER}.c_read (internal_c_pointer, a_field.c_offset))
				when {CE_UTILITIES}.Ce_type_integer_32 then set_integer_32_field (a_field.internal_index, Current, {CE_INTEGER_32_POINTER}.c_read (internal_c_pointer, a_field.c_offset))
				when {CE_UTILITIES}.Ce_type_integer_64 then set_integer_64_field (a_field.internal_index, Current, {CE_INTEGER_64_POINTER}.c_read (internal_c_pointer, a_field.c_offset))
				when {CE_UTILITIES}.Ce_type_natural_8  then set_natural_8_field  (a_field.internal_index, Current, {CE_NATURAL_8_POINTER}.c_read  (internal_c_pointer, a_field.c_offset))
				when {CE_UTILITIES}.Ce_type_natural_16 then set_natural_16_field (a_field.internal_index, Current, {CE_NATURAL_16_POINTER}.c_read (internal_c_pointer, a_field.c_offset))
				when {CE_UTILITIES}.Ce_type_natural_32 then set_natural_32_field (a_field.internal_index, Current, {CE_NATURAL_32_POINTER}.c_read (internal_c_pointer, a_field.c_offset))
				when {CE_UTILITIES}.Ce_type_natural_64 then set_natural_64_field (a_field.internal_index, Current, {CE_NATURAL_64_POINTER}.c_read (internal_c_pointer, a_field.c_offset))
				when {CE_UTILITIES}.Ce_type_real_32    then set_real_32_field    (a_field.internal_index, Current, {CE_REAL_32_POINTER}.c_read    (internal_c_pointer, a_field.c_offset))
				when {CE_UTILITIES}.Ce_type_real_64    then set_real_64_field    (a_field.internal_index, Current, {CE_REAL_64_POINTER}.c_read    (internal_c_pointer, a_field.c_offset))
				when {CE_UTILITIES}.Ce_type_real_96 then
					if attached {REAL_96} field_from_eif_internal (a_field.internal_index, Current) as r96 then
						r96.c_pointer.memory_copy (internal_c_pointer + a_field.c_offset, {CE_UTILITIES}.Real_96_bytes)
					else
						(create {EXCEPTIONS}).raise (field_from_eif_internal (a_field.internal_index, Current).generator)
					end
				when {CE_UTILITIES}.Ce_type_pointer then
					l_tmp_pointer ?= new_instance_of (a_field.eif_type)
					l_tmp_pointer.make_shared ({CE_POINTER_POINTER [ANY]}.c_read (internal_c_pointer, a_field.c_offset))
					set_reference_field (a_field.internal_index, Current, l_tmp_pointer)
				when {CE_UTILITIES}.Ce_type_array then
					-- nothing to do, uses the struct memory
				when {CE_UTILITIES}.Ce_type_class then
					l_tmp_class ?= field_from_eif_internal (a_field.internal_index, Current)
					l_tmp_class.update_class_fields
				when {CE_UTILITIES}.Ce_type_va_list then
					set_reference_field (a_field.internal_index, Current, create {CE_VA_LIST}.make_shared_indirect (internal_c_pointer + a_field.c_offset))
				end
			end
		end

invariant
	internal_c_pointer /= default_pointer implies across structure_info.fields as c all
			(c.item.ce_type = {CE_UTILITIES}.Ce_type_array implies internal_c_pointer + c.item.c_offset = ({CE_ARRAY [ANY]} #? field_from_eif_internal (c.item.internal_index, Current)).c_pointer) and
			(c.item.ce_type = {CE_UTILITIES}.Ce_type_class implies internal_c_pointer + c.item.c_offset = ({CE_CLASS} #? field_from_eif_internal (c.item.internal_index, Current)).c_pointer)
		end
	(structure_info.is_union or structure_info.has_bitfields) implies internal_c_pointer /= default_pointer

end
